/*
The Scriptol library
Home page of the compiler: https://www.scriptol.com

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/


#include <stdlib.h>
#include <stdio.h>
#include <float.h>
#include <stdarg.h>
#include <ctype.h>

#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <algorithm>  
#include <string>

#ifdef UNIX
#include <unistd.h>
#else
#include <io.h>
#endif

#include "text.hpp"
#include "intarr.hpp"
#include "textarr.hpp"
#include "array64.hpp"
#include "dyn.hpp"
#include "array.hpp"
#include "dict.hpp"


/** solarray.cpp
  File of the scriptol array
  */


dyn solarray::nullelement = dyn(false);

/* Constructors */

solarray::solarray() {
  indexer = 0;
}

 /**
   Build an associative array from an array of strings of chars.
   Indexes are generated from 0 to n.
   */

solarray::solarray(char *a[]) { 
    set(a); 
}

solarray::solarray(dyn x) {
    indexer = 0;
    data.push_back(x);
}

solarray::solarray(const solarray &a) {
  data.assign(a.data.begin(), a.data.end());
  indexer = 0;
}

solarray::solarray(solarray *a) {
  data.assign(a->data.begin(), a->data.end());
  indexer = 0;
}

solarray::solarray(std::initializer_list<dyn> il) {
    data.assign(il);
    indexer = 0; 
}


// assign a std::vector

solarray::solarray(std::vector<dyn> &a) { set(a); }
solarray::solarray(std::vector<dyn> *a) { set(a); }

solarray::~solarray() {
  data.clear();
}


void solarray::clear() {
  data.clear();
  indexer = 0;
}

void solarray::set(const solarray &a)  {
  data.assign(a.data.begin(), a.data.end());
  indexer = 0;
}

void solarray::set(solarray *a) {
  data.assign(a->data.begin(), a->data.end());
  indexer = 0;
}

void solarray::set(std::vector<dyn> &a) {
    clear();
    for(int i = 0; i < a.size(); i++) {
        dyn x = a[i];
        data.push_back(x);
    }
    indexer = 0; 
}

void solarray::set(_intArr b) {
  clear();
  for(int i = 0; i < b.size(); i++) {
    push(dyn(b.at(i)));
  }
}

void solarray::set(_arr32 b) {
  clear();
  for(int i = 0; i < b.size(); i++) {
    push(dyn(b.at(i)));
  }
}


void solarray::set(_array64 b) {
  clear();
  for(int i = 0; i < b.size(); i++) {
    push(dyn(b.at(i)));
  }
}


void solarray::set(dyn d) {
  if(d.flag != TYPE_ARRAY) { clear(); push(d); return; }
  set(d.a);
}

void solarray::set(char *x[])
{
  int i = 0;
  this->clear();
  while(x[i] != NULL)  push(dyn(x[i++]));
}

/*
dyn solarray::get(int n)
{
  char buffer[64];
  sprintf(buffer, "%d", n);
  return get(buffer);
}
*/
void solarray::add(const solarray &a)
{
  const solarray *x = &a;
  add((solarray *) x);
}

void solarray::add(solarray *x) {
  for(int i = 0; i < x->size(); i++)  {
      data.push_back(x->data[i]);
  }
}


void solarray::clone(solarray *y)  {
   y->data.assign(data.begin(), data.end());
   y->indexer = 0;
 }

void solarray::pack() {
    return;    
}

//===== OPERATORS

void solarray::operator=(const solarray &a) {
  data.assign(a.data.begin(), a.data.end());
  indexer = 0;
  return;
}

void solarray::operator=(solarray *a)  { set(a); }
void solarray::operator=(dyn a)        { set(a); }
void solarray::operator=(_intArr &a)   { set(a); }
void solarray::operator=(_arr32 &a)    { set(a); }
void solarray::operator=(_array64 &a)  { set(a); }

void solarray::operator=(std::vector<dyn> a) { set(a); };


 solarray solarray::operator+(solarray *x)
 {
   solarray y = solarray(this);
   y.add(x);
   return y;
 }

 solarray solarray::operator+(dyn x)
 {
   solarray y = solarray(this);
   y.push(x);
   return y;
 }

// clone current array minus item in array b

solarray solarray::operator-(solarray *b) {
   solarray y = solarray();
   for(int i = 0; i < size(); i++) {
     for(int j = 0; j  < b->size(); j++)  {
       if(data[i] == &b->data[j]) goto skip;
     }
     y.push(data[i]);
     skip:;
   }
   return y;
};

solarray solarray::operator-(dyn x) {
   solarray y = solarray();
   for(int i = 0; i < size(); i++)  {
     if(data[i] != x)  {
       y.push(data[i]);
     }
   }
   return y;
}

// solarray solarray::operator&(solarray a) { return operator&(a); }
 solarray solarray::operator&(solarray *b)  {
   dyn v;
   int i,j;
   int l = b->size();
   solarray *y = new solarray();

   for(i = 0; i < size(); i++) {
     v.set(data[i]);
     for(j = 0; j  < l; j++) {
       if(v == b->data[j]) {
           y->data.push_back(v);
       }
     }
   }
   return *y;
 };


solarray solarray::operator|(solarray *b) {
  unsigned int i, l = size();
  solarray *y = new solarray();

  for(i = 0; i < l; i++)  {
    y->pushUnic(data[i]);
  }
  y->add(b);
  return y->unique();
}


solarray solarray::operator^(solarray *b) {
  int i;
  solarray *y = new solarray();

  for(i = 0; i < size(); i++)  {
    if(b->findValue(data[i]) == -1)
      y->pushUnic(data[i]);
  }
  for(i = 0; i < b->size(); i++)  {
    if(findValue(b->data[i]) == -1)
      y->pushUnic(b->data[i]);
  }
  return *y;
}

solarray &solarray::operator+=(solarray *x) {
   this->add(x);
   return *this;
}

solarray &solarray::operator+=(dyn x) {
   push(x);
   return *this;
}


void solarray::operator-=(solarray *x)  {
   solarray y = solarray(this);
   set(y.operator-(x));
}

void solarray::operator-=(dyn x) {
   solarray y = solarray(this);
   set(y - x);
 }

 void solarray::operator&=(solarray *b)  {
   solarray y = operator&(b);
   set(y);
 }

 void solarray::operator|=(solarray *b)  {
   solarray y = operator|(b);
   set(y);
 }

 void solarray::operator^=(solarray *b) {
   solarray y = operator^(b);
   set(y);
 }

bool solarray::operator==(solarray *b) {
   int l = size();
   if(b->size() != l) return false;
   for(int i = 0; i < l; i++)  {
     if(data[i] != b->data[i]) return false;
   }
   return true;
 }


 bool  solarray::operator==(bool x)  { 
     return (size() != 0) == x; 
 }

 bool  solarray::operator!=(solarray *a) { 
     return !operator==(a); 
 };
 
 bool  solarray::operator!=(bool x) { 
     return !operator==(x);
 }

 bool solarray::operator<(solarray *a) {
   int l = a->size();
   if(size() < l) return true;
   if(size() > l) return false;
   for(int i = 0; i < l; i++)
   {
     if(data[i] > a->data[i]) return false;
     if(data[i] < a->data[i]) return true;
   }
   return false;
 }

 bool solarray::operator>(solarray *a)  {
   int l = a->size();
   if(size() < l) return false;
   if(size() > l) return true;
   for(int i = 0; i < l; i++)
   {
     if(data[i] > a->data[i]) return true;
     if(data[i] < a->data[i]) return false;
   }
   return false;
 }

bool solarray::operator!() { 
    return size() == 0; 
}

solarray &solarray::operator,(dyn x) { 
    push(x); 
    return *this; 
}

// utils

 bool solarray::in(const dyn &d)  {
    for(unsigned int i = 0; i < data.size(); i++)  {
        if(data[i] == d) return true;
    }
    return false;
 }

 bool solarray::in(dyn *d)  {
    for(unsigned int i = 0; i < data.size(); i++)  {
        if(data[i] == d) return true;
    }
    return false;
 }

signed int solarray::findValue(dyn *d) { 
    return findValue(*d); 
}

signed int solarray::findValue(const dyn &d) {
  for(int i = 0; i < size(); i++)  {
   if(data[i] == d) return i;
  }
  return -1;
}

/*
  Find last value
*/

signed int solarray::findLastValue(dyn *d) { 
    return findLastValue(*d); 
}

signed int solarray::findLastValue(const dyn &d) {
  for(int i = size() - 1; i >= 0; i--)  {
   if(data[i] == d) return i;
  }
  return -1;
}

 //===== iterator

 void solarray::reset() { 
     indexer = 0; 
 }
 
 bool solarray::outOfBound() { 
     return indexer >= size(); 
 }

 bool solarray::bound()  { 
     return(++indexer < size()); 
 }
 
 bool solarray::inBound() {
  if(indexer < size()) { 
      return true;
  }
  return false;
 }
 
 dyn solarray::inc()  {
  if(++indexer < size()) { 
      return data[indexer]; 
  }
  indexer = size();
  return dyn(false);
 }

 dyn solarray::dec() {
  if(--indexer >= 0) return data[indexer];
  indexer = 0;
  return dyn(false);
 }

 dyn solarray::current() { 
     return value(indexer); 
 }

 dyn solarray::value() { 
     return value(indexer); 
 }

 dyn solarray::value(int i) {
   if(i >=0 && i < size())  return data[i];
   return dyn(false);
  }

text solarray::key() {
   return text(indexer);
}

dyn solarray::begin() {
   indexer = 0;
   if(data.size() > 0) return data.front();
   return dyn(false);
}

dyn solarray::end()   {
   indexer = 1;
   if(data.size() > 0)  {
     indexer = (int) data.size() -1;
     return data.back();
   }
   return dyn(false);
}

//===== stack

void solarray::push(const dyn &a) {
    data.push_back(a);
}

 void solarray::push(const char *s) {
	dyn *d = new dyn(s);
	data.push_back(*d);
 }

 void solarray::push(dyn *a) {
    data.push_back(*a);
 }


void solarray::pushUnic(dyn v) {
     if(findValue(v) == -1) {
       data.push_back(v);
     }
}

dyn solarray::pop() {
   dyn x = dyn(false);
   indexer = 0;
   if(!empty())  {
     x.set(data.back());
     data.pop_back();
   }
   return x;
}

dyn solarray::shift() {
	if(empty())  return dyn(false);
	else {
		dyn *x = &data[0];
		erase(0);
		return *x;
	}
}

void solarray::unshift(dyn a)
{
  data.insert(data.begin(), a);
}

void solarray::insert(int n, dyn d) {
  if(n < 0) n = size() + n;
  if(n < 0) return;
  if(n >= size())
  {
    if(d.flag == TYPE_ARRAY) add(d.a);
    else  push(d);
    return;
  }

  if(d.flag == TYPE_ARRAY) {
   data.insert(data.begin() + n, d.a->data.begin(), d.a->data.end());
  }
  else  {
   data.insert(data.begin() + n, d);
  }
}

//===== RANGE

 // erase from first to last included
 void solarray::erase(int first, int last) {
    data.erase(data.begin() + first, data.begin() + last + 1);
 }

 void solarray::erase(int pos)  {
    data.erase(data.begin() + pos);
 }


 solarray solarray::slice(int f, int l, int incl)  {
   solarray x;

   if(empty()) goto bye;
   l += incl;
   l--;
   if(l < 0) l += size();
   if(l < 0) goto bye;
   if(l >= size()) l = size() - 1;

   if(f < 0) f += size();
   if(f < 0) f = 0;
   if(l < f) goto bye;
   if(f >= size()) goto bye;

   x.data.assign(data.begin() + f, data.begin() + l + 1);
bye:
   return x;
 }

 void solarray::splice(int b, int e)  {
   if(b < 0) b = size() + b;
   if(b < 0) b = 0;
   if(e < 0) e = size() + e;
   if(e < 0) return;
   if(e >= size()) e = size() - 1;
   if(e < b) return;
   erase(b, e);
 }

 void solarray::splice(int b, int e, int inc)  {
   splice(b, e + inc - 1);
 }


 void solarray::splice(int f)  { 
     erase(f); 
 }

 void solarray::splice(int i, int last, solarray a, int incl) { 
     splice(i, last, &a, incl); 
 }
 
 void solarray::splice(int i, int last, solarray *a, int incl) {
  if(empty()) { operator=(a); return; }
  if(last < 0) last = size() - last;
  if(i < 0) i += size();
  if(i < 0) i = 0;
  last += incl;
  last--;
  if(last >= size()) last = size() - 1;
  if(i >= size()) { add(a); return; }
  if(last < i) return;
  if(last < 0) return;
  if((last + 1) == i)  {
    if(!a->empty()) insert(i, dyn(a));
    return;
  }
  if(a->empty()) { 
      erase(i, last); return; 
  }

  int j = 0;
  while((i <= last) && (j < a->size()))  {
    data[i].set(a->data[j]);
    i++;
    j++;
  }
  if(last >= i) erase(i, last);
  if(j < a->size())  {
   data.insert(data.begin() + i, a->data.begin() + j, a->data.end());
  }
 }


 void solarray::splice(int i, int last, dyn a, int incl)  {
  if(i < 0) i = size() - i;
  if(i < 0) i = 0;
  if(last < 0) last = size() - last;
  last += incl;
  last --;
  if(last >= size()) last = size() - 1;
  if(last < i) return;
  if(last < 0) return;

  if(i >= size()) { push(a); return; }
  if(last == i)   { insert(i, a); }

  data[i].set(a);
  erase(i + 1, last);
 }

//===== TRANSFORM

/*
  if data is a not a number, search if exists further
  if yes, wait for the next occurence to push
  if not a number push it providing not already pushed
*/

solarray solarray::unique() {
  int i, l = size();
  solarray y = solarray();

  for(i = 0; i < l; i++)  {
    if(y.findValue(data[i]) == -1)
      y.push(data[i]);
  }
  return y;
}


void solarray::swapData(int a, int b) {
  dyn x;
  x.set(data[a]);
  data[a].set(data[b]);
  data[b].set(x);
}

void solarray::swapElem(int a, int b)
{
  dyn x;
  x.set(data[a]);
  data[a].set(data[b]);
  data[b].set(x);
}

void solarray::dynSort(int first,int last) {
 int f=first;
 int l=last;
 dyn item;
 item.set(data[(f + l) / 2]);

 do {
  while(data[f] < item) f++;
  while(item < data[l]) l--;
  if(f <= l) {
   swapElem(f, l);
   f++;
   l--;
  }
 } while(f <= l);

 if(first < l)  dynSort(first, l);
 if(f < last)   dynSort(f, last);
 return;
}

void solarray::dynRSort(int first, int last) {
 dynSort(first, last);
 reverse();
 return;
}


void solarray::sort(int x)  { 
    x=0; 
    dynSort(0, size() - 1); 
}

void solarray::rSort(int x) { 
    x=0; 
    dynRSort(0, size()-1); 
}

solarray solarray::keys() {
  solarray x = solarray();
  for(unsigned int i = 0; i < size(); i++) x.push(dyn(i));
  return x;
}

solarray solarray::values() {
  solarray x = solarray();
  x.data.assign(data.begin(), data.end());
  return x;
}

dyn solarray::min() {
  double m = DBL_MAX;
  dyn d;
  double x = 0;
  for(int i = 0; i < size(); i++)
  {
    d.set(data[i]);
    if(d.flag == TYPE_NUMBER) x = d.toReal();
    else
     if(d.flag == TYPE_TEXT)  x = d.t->toNumber();
    if(x < m) m = x;
  }
  return dyn(m);
}

dyn solarray::max() {
  double m = 0;
  dyn d;
  double x = 0;
  for(int i = 0; i < size(); i++)
  {
    d.set(data[i]);
    if(d.flag == TYPE_NUMBER) x = d.toReal();
    else
     if(d.flag == TYPE_TEXT)  x = d.t->toNumber();
    if(x > m) m = x;
  }
  return dyn(m);
}

/*
  Functional map method
  */

solarray solarray::map(bool (*f)(dyn)) {
  int s = data.size();
  solarray temp;
  for(int i = 0; i < s; i++)  {
    if(f(data[i]))
      temp.push(data[i]);
  }
  return temp;
}

double solarray::sum() {
  double total = 0;
  dyn d;
  for(int i = 0; i < size(); i++)  {
    d.set(data[i]);
    if(d.flag == TYPE_NUMBER) total += d.toReal();
    else
     if(d.flag == TYPE_TEXT)  {
       total += d.t->toNumber();
     }
  }
  return total;
}

//===== OPERATIONS

text solarray::join(const char *sep) {
  int lsep, si, sum, offset;
  int i;

  lsep = strlen(sep);
  si = size();
  sum = 0;
  offset = 0;

  for(i = 0; i < si; i++)  {
    if(data[i].flag == TYPE_TEXT)   {
     sum+= data[i].t->length();
     sum+= lsep;
    }
  }

  char *nstr = (char *) new char[sum+1];
  for(i = 0; i < si; i++)  {
   if(data[i].flag == TYPE_TEXT)  {
     if(offset > 0) {
      strcpy(nstr + offset, sep);
      offset += lsep;
     }
     strcpy(nstr + offset, data[i].t->data);
     offset += data[i].t->length();
     nstr[offset] = '\0';
   }
  }
  nstr[offset] = '\0';
  return text(nstr);
}

solarray solarray::reverse() {
  solarray y = solarray();
  for(int i = size() - 1; i >= 0; i--)  {
    y.data.push_back(data[i]);
  }
  return y;
}


std::ostream& operator<< (std::ostream& out, solarray *d)
{
  return out << d->toText().data;
}


void solarray::display() { display(0); }

void solarray::display(int tab) {
 int l = size();
 dyn x;
 std::cout << "solarray(\n";
 for(int i = 0; i < l; i++)
 {
   //std::cout << "flag=" << data[i].flag << " " << data[i].toStr() << std::endl;
   for(int j = 0; j < tab; j++) std::cout << "     ";
   std::cout << "   " << i;
   std::cout << " : ";
   x.set(data[i]);
   if(x.flag == TYPE_ARRAY) x.a->display(tab + 1);
   else
    std::cout << x.toStr() << std::endl;
 }
 for(int j = 0; j < tab; j++) std::cout << "     ";
 std::cout << ")\n";
}

// Convert solarray to string

char *solarray::toString(const char *s) {
	int l = size();
	text *t = new text();
	for(int i = 0; i < l; i++) 	{
		if(i>0) t->cat(s);
		t->cat(data[i].toStr());
	}
	return t->data;
}

text solarray::toText(const char *s) {
	int l = size();
	text *t = new text();
	for(int i = 0; i < l; i++) 	{
		if(i>0) t->cat(s);
		t->cat(data[i].toStr());
	}
	return t;
}


std::ostream& operator<< (std::ostream& out, solarray d) {
  return out << d.toText().data;
}

void solarray::load(const char *name) {
  std::ifstream input(name);
  this->clear();
	for(std::string line; getline(input, line); ) {
	  push(line.c_str());
	}
	input.close();
	return;
}

bool solarray::store(const char * name)
{
    std::ofstream f;
    f.open(name);
	for(int i = 0; i < size(); i++)	{
        f << data[i].toStr() << "\n";
	}
	f.close();
	return true;
}

solarray *solarray::to(int n) {
    if(n >= size())  {
        n = size();
        data.push_back( new solarray() );
   }
   return data[n].toArray();
}

dyn &solarray::operator[](int n)  {
    if(n >= size())  {
        int i = size();
        while(i <= n) {
            data.push_back( dyn(false) );
            i++;
        }  
   }
   return *(data.begin() + n);
 }

dyn &solarray::at(int n)  {
   if(n >= size())  {
     dyn::error = true;
     return nullelement;
   }
   return data[n];
 }
 

dyn &solarray::at(dyn d) { 
    return at(d.toInt());
} 

void solarray::setAt(int n, dyn *a) {
   if(n < size())  {
     data[n].set(*a);
     return;
   }
   push(*a);
 }

void solarray::setAt(int n, dyn a)  {
   if(n < size()) {
     data[n].set(a);
     return;
   }
   push(a);
 }


void solarray::boundError(const text &t) {
  std::cout << "Out of boundary " << t << std::endl;
}
