/*
The Scriptol library
Home page of the compiler: https://www.scriptol.com

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/


#include <stdlib.h>
#include <stdio.h>
#include <float.h>
#include <stdarg.h>
#include <limits.h>

#include <iostream>
#include <vector>

#include "text.hpp"
#include "array64.hpp"

/** \file arr64.cpp
  File of the typed array 64 bits library
  */

/** \class _array64
   Base class for typed array of 32 bits objects
   including integers, string of chars, pointers on objects...
 */


int _default_capacity64 = 16;

//=====  Constructors

/**
   Basic constructor
  */
  
_array64::_array64()           { set(); }
_array64::_array64(int t)      { set(); type = t; }
_array64::_array64(_array64 *a)  { set(a); }
_array64::_array64(const _array64 &a) { set(a); }

_array64::_array64(int t, int c)
{
  set(c);
  type = t;
}

_array64::_array64(int t, double n)
{
  set();
  type = t;
  push(n);
}

_array64::_array64(int t, int n, double v)
{
  set(n);
  type = t;
  push(v);
}


//===== Utilities

void _array64::clear() {
  set(type, _default_capacity64);
}

void _array64::resize(int i) {
    if(i >= len)  {
        if(i >= capacity)  {
            capacity = i + 16;
            data = (double *) realloc(data, capacity * sizeof(double));
        }
        len = i + 1;
    }
}



char *_array64::toStr(int n)
{
  char *buffer = (char *) new char[32];
  sprintf(buffer, "%d", n);
  return buffer;
}

char *_array64::toStr()
{
   text t = toText();
   return t.data;
}

// comparing two items of the array

int _array64::compare(double a, double b)
{
   double di, dj;

   switch(type)
   {
     case TYPE_NATURAL:
     case TYPE_NUMBER:
          di = (double) a;
          dj = (double) b;
          if(di < dj) return -1;
          if(di > dj) return 1;
          return 0;

     default:
          return 0;

   }
}


void _array64::pack() {
  int i, j;
  j = 1;
  for(i = 0; i < len; i++) {
    if(data[i] == (double) NULL) {
      while(data[j] == (double) NULL) j++;
      if(j >= len) {
        len = i;
        return;
      }
      data[i] = data[j];
    }
    j++;
  }
  len = i;
  return;
}

void _array64::set() { set(_default_capacity64); }
void _array64::set(int si) {
    data = (double*) new char[si * sizeof(double)];
    capacity = si;
    type = 0;
    len = 0;
    indexer = 0;
}

void _array64::set(int ty, int si) {
   set(si);
   type = ty;
}


void _array64::set(const _array64 &model) {
    set(model.type, model.capacity);
    for(int i  = 0; i < model.len; i++)  {
      push(model.data[i]);
    }
}

void _array64::set(_array64 *model) {
    set(model->type, model->capacity);
    for(int i  = 0; i < model->len; i++) {
      push(model->data[i]);
    }
}



//===== GET

double * _array64::operator[](int i)  {
   if(i >= len)  {
     resize(i);
     data[i] = (double) NULL;
   }
   return & data[i];
 }

double * _array64::operator[](const char *s)  {
   int k = atoi(s);
   if(k >= len) {
    resize(k);
    data[k] = (double) NULL;
   }
   return & data[k];
 }


double  _array64::at(int n) {
   if(n < len) return data[n];
   return (double) NULL;
 }

double _array64::at(text t) { return at(t.data); }
double _array64::at(const char *a) {
    if(text::isNumber(a))  {
        int n = atoi(a);
        if(n < len)  return data[n];
    }
    return (double) NULL;
}


void _array64::setAt(int i, double a) {
   if(i < len) data[i] = a;
   else push(a);
 }


void _array64::add(_array64 x)
{
  int l = x.size();
  for(int i = 0; i < l; i++)
  {
     push(x.data[i]);
  }
}


void _array64::clone(_array64 *y)
 {
   y->set(*this);
 }


//===== OPERATORS

void _array64::operator=(const _array64 &a)  { set(a); }
void _array64::operator=(_array64 *a) { set(a); }
void _array64::operator=(double a)   { set(); push((double) a); }
void _array64::operator=(natural n)  { set(); push((double) n); }
void _array64::operator=(int n)  { set(); push((double) n); }


_array64 _array64::operator+(_array64 x)
 {
   _array64 y = _array64(this);
   y.add(x);
   return y;
 }

 _array64 _array64::operator+(double x)
 {
   _array64 y = _array64(this);
   y.push(x);
   return y;
 }


_array64 _array64::operator-(_array64 b)
{
   int i,j;
   int l = b.size();
   _array64 y = _array64(type, len);
   double x;

   for(i = 0; i < len; i++)
   {
     x = data[i];
     //std::cout << (int)  x << std::endl;
     for(j = 0; j  < l; j++)
     {
       if(compare(x ,b.data[j]) == 0)
       {
        //std::cout << "skip " << (int) x << std::endl;
        goto skip;
       }
     }
     //std::cout << "push " << (int)  x << std::endl;
     y.push(x);
skip:;
   }
   return y;
};

_array64 _array64::operator-(double x)
{
   _array64 y = _array64(type, len);
   for(int i = 0; i < len; i++)
   {
     if(compare(data[i] , x))  y.push(data[i]);
   }
   return y;
}

 _array64 _array64::operator&(_array64 b)
 {
   double v;
   int i,j;
   int l = b.size();
   _array64 y = _array64(type, len);

   for(i = 0; i < size(); i++)
   {
     v = data[i];
     for(j = 0; j  < l; j++)
     {
       if(!compare(v , b.data[j]))
           y.push(v);
     }
   }
   return y;
 };

_array64 _array64::operator&(double v)
{
  _array64 y = _array64(type, 16);
  if(findValue(v) > -1) y.push(v);
  return y;
}


_array64  _array64::operator|(_array64 b)
{
  int i, l = b.size();
  _array64 y = _array64(type, len + l);

  for(i = 0; i < len; i++)
    y.pushUnic(data[i]);

  for(i = 0; i < l; i++)
    y.pushUnic(b.data[i]);

  return y;
}

_array64 _array64::operator|(double b)
 {
   _array64 y = _array64(this);
   y.pushUnic(b);
   return y;
 }


_array64 _array64::operator^(_array64 b)
{
  int i;
  _array64 y = _array64(type, len);

  for(i = 0; i < len; i++)
  {
    if(b.findValue(data[i]) == -1)
      y.pushUnic(data[i]);
  }
  for(i = 0; i < b.size(); i++)
  {
    if(findValue(b.data[i]) == -1)
      y.pushUnic(b.data[i]);
  }
  return y;
}


 void _array64::operator-=(_array64 x)
 {
   _array64 y = _array64(this);
   set(y.operator-(x));
 }

 void _array64::operator-=(double x)
 {
   _array64 y = _array64(this);
   set(y.operator-(x));
 }

void _array64::operator&=(_array64 b)
 {
   _array64 y = operator&(b);
   set(&y);
 }

void _array64::operator&=(double v)
 {
   int flag = findValue(v) > -1;
   clear();
   if(flag) push(v);
 }


void _array64::operator|=(_array64 b)
 {
   _array64  y = operator|(b);
   set(&y);
 }

void _array64::operator|=(double v)
{
   pushUnic(v);
}

void _array64::operator^=(_array64 b)
 {
   _array64  y = operator^(b);
   set(&y);
 }


bool _array64::operator==(_array64 b)
 {
   if(b.size() != len) return false;
   for(int i = 0; i < len; i++)
   {
     if(compare(data[i] , b.data[i])) return false;
   }
   return true;
 }

 bool  _array64::operator==(bool x)  { return (size() != 0) == x; }

 bool _array64::operator<(_array64 a)
 {
   int l = a.size();
   if(len < l) return true;
   if(len > l) return false;
   for(int i = 0; i < l; i++)
   {
     int r = compare(data[i] , a.data[i]);
     if (r > 0) return false;
     if (r < 0) return true;
   }
   return false;
 }

 bool _array64::operator>(_array64 a)
 {
   int l = a.size();
   if(len < l) return false;
   if(len > l) return true;
   for(int i = 0; i < l; i++)
   {
     int r = compare(data[i] , a.data[i]);
     if (r < 0) return false;
     if (r > 0) return true;
   }
   return false;
 }

bool _array64::operator!() { return len == 0; }



//===== util

 bool _array64::in(double d)
 {
  for(int i = 0; i < size(); i++)
  {
   if(!compare(data[i] , d)) return true;
  }
  return false;
 }

/* is the textual key already in dict? */
signed int _array64::hasStrKey(const text &s)
{
  if(text::isNumber(s.data))
  {
    int i = atoi(s.data);
    return len - i;
  }
  return -1;
}

signed int _array64::findKey(const text &s)
{
  if(text::isNumber(s.data))
  {
    return atoi(s.data);
  }
  return -1;
}

signed int _array64::findKey(int n) { return n; }

signed int _array64::findValue(double d)
{
  for(int i = 0; i < size(); i++)
  {
   if(!compare(data[i] , d)) return i;
  }
  return -1;
}

void _array64::display() { display(0); }
void _array64::display(int tab) {
    double x;

    std::cout << "Array of ";
    switch(type) {
        case TYPE_REAL:  std::cout << "real(\n";
            break;
        case TYPE_NUMBER:  std::cout << "number(\n";
            break;
        case TYPE_NATURAL:  std::cout << "natural(\n";
            break;
        default:  std::cout << "object(\n";
            break;
    }
 
    for(int i = 0; i < len; i++) {
        for(int j = 0; j < tab; j++) std::cout << "     ";
        std::cout << "   " << i;
        std::cout << " : ";
        x = data[i];
        switch(type) {
            case TYPE_NATURAL:
                std::cout << (natural) x;
                break;
            default:
                std::cout << x;
                break;
        }
        std::cout << std::endl;
    }
    for(int j = 0; j < tab; j++) std::cout << "     ";
    std::cout << ")\n";
}

char *_array64::toString(const char *s) {
	int l = size();
	text *t = new text();
    switch(type)  {
        case TYPE_REAL:     t->cat("real(");
            break;
        case TYPE_NUMBER:   t->cat("number(");
            break;
        case TYPE_NATURAL:  t->cat("natural(");
            break;
        default:            t->cat("object(");
            break;
    }	
	for(int i = 0; i < l; i++) 	{
		if(i > 0) t->cat(s);
		t->cat(toStr(data[i]));
	}
	t->cat(")");
	return t->data;
}

text _array64::toText(const char *sep) {
    return toString(sep);
}


 //===== iterator

 void _array64::reset() { indexer = 0; }
 bool _array64::outOfBound() { return indexer >= size(); }
 bool _array64::bound()      { return(++indexer < size()); }
 
  bool _array64::inBound() {
  if(indexer < size()) { 
      return true;
  }
  return false;
 }

 double _array64::inc()  {
  if(++indexer < size()) { return data[indexer]; }
  indexer = size();
  return (double) 0;
 }

 double _array64::dec() {
  if(--indexer >= 0) return data[indexer];
  indexer = 0;
  return (double) 0;
 }

 double _array64::current() { return value(indexer); }
 double _array64::value() { return value(indexer); }
 double _array64::value(int i)
 {
   if(i >=0 && i < size())  return data[i];
   return (double) 0;
 }

 int _array64::key() {
   if(indexer >= 0 && indexer < size())  return indexer;
   return -1;
  }

 int _array64::key(int i)
 {
   if(i >= 0 && i < size())  return i;
   return -1;
 }

 int _array64::index()
 {
   if(indexer >=0 && indexer < size()) return indexer;
   return -1;
 }

double _array64::begin()
 {
   indexer = 0;
   if(size() > 0) return data[ 0 ];
   return (double) NULL;
}

double _array64::end()
{
   indexer = 1;
   if(size() > 0)
   {
     indexer = size() -1;
     return data[indexer];
   }
   return (double) NULL;
}

//===== stack


 void _array64::push(double a) {
    resize(len);      
    data[len-1] = a;
 }

 void _array64::pushUnic(double v) {
     if(findValue(v) == -1)  {
       push(v);
     }
 }

 double _array64::pop() {
    indexer = 0;
    if(empty()) return (double) NULL;
    len--;
    return data[len];
 }

double _array64::shift() {
    if(empty()) return (double) NULL;
    double x = data[0];
    erase(0);
    return x;
}

void _array64::unshift(double a) {
    insert(0, a);
}

void _array64::insert(int n, double d) {
    if(n < 0) n += len;
    if(n < 0) return;
    if(n >= len)  {
        push(d);
        return;
    }

    double x = data[len - 1];
    for(int i = len - 1; i > n; i--)  {
        data[i] = data[i - 1];
    }
    data[n] = d;
    push(x);
}

void _array64::insert(int n, _array64 d) {
    if(n < 0) n += len;
    if(n < 0) return;
    if(n >= len)  {
        add(d);
        return;
    }

    int si = len + d.size();

    _array64 x = _array64(type, si);
    if(n > 0) x.add(slice(0, n, 0));
    x.add(d);
    x.add(slice(n));
    set(x);
}

//===== RANGE


 // erase from first to last included
 
 void _array64::erase(int first, int last)  {
   int k, l;

   if(first > len) return;
   if(last >= len)  {
      len = first + 1;
      return;
   }

   l = len;
   int nl = len - (last - first + 1);
   if(nl < 0) nl = 0;
   k = last;
   while(first < len)  {
     k++;
     if(k < l)  {
       data[first] = data[k];
     }
     else  {
       data[first] = (double) NULL;
     }
     first++;
   }
   len = nl;
 }

 void _array64::erase(int pos)  {
    if(pos >= len) return;
    len--;
    while(pos < len)  {
       data[pos] = data[pos + 1];
       pos++;
    }
    data[len] = (double) NULL;
    return;
 }

 _array64 _array64::slice(int f, int l, int incl)
 {
   _array64 x;

   l += incl;
   l--;
   if(l < 0) l += len;
   if(l < 0) goto bye;
   if(l >= len) l = len - 1;

   if(f < 0) f += len;
   if(f < 0) f = 0;
   if(l < f) goto bye;
   if(f >= len) goto bye;

   x = _array64(type, l - f + 1);

   if(empty()) goto bye;
   //std::cout << f << " - " << l << std::endl;
   //x.display();

   while(f <= l)
   {
     //std::cout << f << ") " << (int) data[f] << std::endl;
     x.push(data[f++]);
   }
bye:
   return x;
 }

 void _array64::splice(int b, int e)
 {
   if(b < 0) b += len;
   if(b < 0) b = 0;
   if(e < 0) e += len;
   if(e < 0) return;
   if(e >= len) e = len - 1;
   if(e < b) return;
   //std::cout << b << " - " << e << std::endl;
   erase(b, e);
 }

 void _array64::splice(int b, int e, int inc)
 {
   splice(b, e + inc - 1);
 }

 void _array64::splice(int f)  { erase(f); }

 void _array64::splice(int i, int last, _array64 a, int incl)
 {
  if(empty())   { operator=(a); return; }
  if(a.empty()) { erase(i, last + incl - 1); return; }
  if(last < 0) last += len;
  if(i < 0) i += len;
  if(i < 0) i = 0;
  last += incl;
  last--;
  if(last >= len) last = len - 1;
  //cout << i << " " << last << " " << size() << std::endl;
  if(i >= len)
  {
    add(a);
    return;
  }
  if(last < i) return;
  if(last < 0) return;
  if((last + 1) == i)
  {
    insert(i, a);       // providing not a.empty() as tested above
    return;
  }
  if(a.empty())
  {
    erase(i, last);
    return;
  }

  int j = 0;
  while((i <= last) && (j < a.size()))
  {
    data[i] = a.data[j];
    i++;
    j++;
  }
  if(last >= i) erase(i, last);
  if(j < a.size())
  {
    insert(i, a.slice(j));
  }
 }


 void _array64::splice(int i, int last, double a, int incl)
 {
  if(i < 0) i += len;
  if(i < 0) i = 0;
  if(last < 0) last += len;
  last += incl;
  last --;
  if(last >= len) last = len - 1;
  if(last < i) return;
  if(last < 0) return;

  if(i >= len)  { push(a); return; }

  data[i] = a;
  if(last > i)  {
    erase(i + 1, last);
  }
 }

//===== TRANSFORM

/*
  if data is a not a number, search if exists further
  if yes, wait for the next occurence to push
  if not a number push it providing not already pushed
*/

_array64 _array64::unique() {
    _array64 y = _array64(type, len);
    for(int i = 0; i < len; i++) {
        y.pushUnic(data[i]);
    }
    return y;
}


void _array64::swapData(int a, int b) {
  double x;
  x = data[a];
  data[a] = data[b];
  data[b] = x;
}


void _array64::dynSort(int first,int last) {
 int f=first;
 int l=last;
 double item = data[(f + l) / 2];

 do  {
   while(data[f] < item) f++;
   while(item < data[l]) l--;
   if(f <= l)  {
     swapData(f, l);
     f++;
     l--;
   }
 } while(f <= l);

 if(first < l)  dynSort(first, l);
 if(f < last)   dynSort(f, last);
 return;
}

void _array64::sort(int f)  { dynSort(0, len - 1); }
void _array64::aSort(int f) { dynSort(0, len - 1); }


_array64  _array64::keys() {
  _array64 x = _array64(TYPE_INT, len);
  for(int i = 0; i < len; i++)  {
    x.push((double) i);
  }
  return x;
}

_array64 _array64::values() {
  return *this;
}

//===== UTILITIES

double _array64::min() {
  double m = INT_MAX;
  double x = 0;

  for(int i = 0; i < len; i++)  {
    x = (double) data[i];
    if(x < m) m = x;
  }
  return m;
}

double _array64::max() {
  double m = INT_MIN;
  double x = 0;

  for(int i = 0; i < len; i++)  {
    x = data[i];
    if(x > m) m = x;
  }
  return m;
}


double _array64::sum() {
  double m = 0;
  for(int i = 0; i < len; i++)  {
      m += data[i];
  }
  return m;
}


text _array64::join(const char *sep) {
  int i;
  int si = size();
  text nstr = "";

  for(i = 0; i < si; i++)  {
     nstr += text(data[i]);
  }
  return nstr;
}

_array64 _array64::reverse() {
  _array64 y = _array64(type, len);
  for(int i = len - 1; i >= 0; i--)
   y.push(data[i]);
  return y;
}

std::ostream& operator<< (std::ostream& out, _array64 &d)
{
  return out << d.toStr();
}

void _array64::load(const char *name)
{
  FILE *f;
  char BUFFER[65536];
  char *ptr;

  f = fopen(name, "r");
  if(f == NULL) return;
  set();

  while(!feof(f))
  {
    ptr = fgets(BUFFER, 65536, f);
    if(ptr == NULL) break;
    push(text::toNumber(BUFFER));
  }

  fclose(f);
  return;
}

bool _array64::store(const char * name)
{
  FILE *f;
  bool res = false;
  text t = new text();

  f = fopen(name, "w");
  if(f == NULL) goto bye;

  for(int i = 0; i < size(); i++)
  {
    t.set(data[i]);
    fputs(t.data, f);
  }
  fclose(f);
  res = true;
bye:
  return res;
}


void _array64::boundError(const text &t)
{
  std::cout << "Out of boundary " << t << std::endl;
}

