/*
The Scriptol library
Home page of the compiler: https://www.scriptol.com

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.com.
*/



using namespace std;

class dict;


/**
   Attribute of an element
  */

class xattribute 
{
  public:
  text name = "";
  text value = "";

  xattribute(const text &, const text &);
  void set(const text &, const text &);
};




/**
   Xml element
  */

class xelement
{
  public:
    char * name;            // tag name
    vector<xattribute *> attributes;
    char * data;
    int index;
    bool isFound;

    xelement();
    xelement(xelement *);
    void set(xelement *);
    void assign(xelement *);

    xelement(const char *);
    void setName(const char *);
    inline char *getName() { return name; }
    void addAttribute(xattribute *);
    void addAttribute(const text &, const text &);
    bool hasAttribute(const char *name);
    xattribute *getAttribute();
      
    inline int size() { return (int) attributes.size(); } 

  /**
    Returns the value for an attribute
    Argument: the name of the attribute
    */
    const char *getValue(const char *);

  /**
    Change the value for an attribute
    Argument: the name of the attribute and the new value
    */
    void setValue(const char *, const char *);

  /**
    Set data
    */

    void setData(const char *);

  /**
    Get data
    */
  
    inline bool hasData() { return strlen(data) > 0; }
    inline const char *getData() {  return data; }


    xelement *begin();
    const char *nextName();
    const char *nextValue();
    bool found();
    xelement *inc();
    void display();
};


/**
   Embedded script
  */

class xscript 
{
  vector<text *> lines;
  vector<int> tabs;
  int index;
  int tab;

  public:
  xscript();
  xscript(xscript *);
  void set(xscript *);

  inline int size() { return (int) lines.size(); }
  void push(text *);
  void push(const string &);
  void reset();
  const char *next();
  int getTab();
  bool again();
  void load(const char *);
};

/**
   Comment line
  */

class xcomment 
{
  public:
  text data;

  xcomment();
  xcomment(xcomment *);
  void set(xcomment *);

  inline void set(const char *s)         { data.set(s); }
  inline void operator=(const char *s)   { data.set(s); }
  inline void operator=(const string &s) { data.set(s.c_str()); }
};

/**
   Node of the dom tree
  */

class dom;

class xnode
{
  public:

  int type;
  int next;
  int child;
  int parent;
  int lastChild;
  dom *owner;

  enum { ELEMENT, SCRIPT, COMMENT };
  
  union
  {
   xelement *element;
   xscript  *statement;
   xcomment *comment;
  };

  /**
    Constructor
    Arguments: optionally another xnode and its owner
    */

  xnode();
  xnode(xnode *);
  xnode(dom *, xnode *);
  void set(xelement *);
  void set(xscript *);
  void set(xcomment *);

  /**
    Set
    Copies the type and the content from a xnode into this one
    other fields are unchanged
    or empty (usually follows a call to the constructor)
    */
  
  char *getNext();

  void set(xnode *);       // copy content and type
};


/**
  The xml tree builder, reader and modifier
  */

class dom 
{
 private:
	vector<text *> header;
	FILE *domfile;
	bool fileFlag;
	bool htmlFlag;
	bool isFound;
	text name;
	int mode;       // display or save mode
	enum { ELEMENT, SCRIPT, COMMENT };

	xnode *node;
	union {
		xelement *element;
		xscript *statement;
		xcomment *comment;
	};

	int index;      // current node number
	int top;
	int tab;


 public:
	vector<xnode *> nodes;	 
 
	static const int XML;
	static const int XSL;
	static const int SVG;
	static const int XUL;
	static const int LIGHT;        // displays light xml
	static const int TREE;         // displays as a tree

	static const int LIBXML;       // use libxml as parser
	static const int EXPAT;        // use expat as parszer

	dom();
	dom(const char *);
	~dom();

	inline void setHtml() { htmlFlag = true; }
	inline int size()   { return (int) nodes.size(); }
	inline bool empty() { return(nodes.size() < 2); }
	inline void setName(const char *s) { element->setName(s); }
	inline char *getName() { return element->name; }
	inline bool found() { return isFound; }

	
 /**
   Assigning another document: fill the tree from its tree
   */

	inline void set(dom *x) { assign(x); }
	void assign(dom *);


	const char *getElementName();

 /**
   Erase the tree
   */

	void clear();
 
	bool isElder(int); 		// is the node an elder, has no previous node?

/**
  Adds a successor ot node currently pointed out (by index)
  */
	int addNext(xnode *, int);

/**
  Adds a child node to node currently pointed out (by index)
  */
	int addChild(xnode *, int);

	int addNext(xelement *);
	int addNext(int, xelement *);
	int addChild(xelement *);
	int addChild(int, xelement *);

	int addNext(xscript *);
	int addChild(xscript *);

	int addNext(xcomment *);
	int addChild(xcomment *);

 /**
   Adds an attribute to current element
   */

	int addAttribute(xattribute *);
	int addAttribute(int, xattribute *);
	int addAttribute(const char *, const char *);


	static const char *tabtxt;
	void indent();

 // display the starting of a node
	void openTag(xelement *, int);

 // display the ending of a node
	void closeTag(xelement *, int);

 // display the content of a node

	void outData(const char *, int);
	void dump();
	void subout(int, int);
	void display(int = XML);


 /**
    Sends a line to screen or file
  */

	void send(const char *);
	void send(char);
	void sendText(const char *);


 /**
    Adds a line to the document
   */

	void addComment(const char *);

 /**
    Adds a line to the header
   */
	void addHeader(const char *);


 /**
    Adds a script file to the document
   */
	void addScriptChild(const char *);
	void addScriptNext(const char *);

 /**
    Adds a script to the document
   */
	void addScript(xscript *);

 /**
    Adds a line to the current script
   */
	void addLine(const string &);

 /**
    Sends header to output
   */
	void outHeader();


 /**
    Set 'index' to the root
    Returns a pointer of the instance of dom, for chaining
   */

 dom *start();

 /**
    Points out the node in the path. Nothing else.
   */

	inline dom *at() { return this; }

 /**
    Returns a pointer on the instance of this class for chaining
    The attribute 'index' must point out a node or the root
    'at' returns the first child node
    the element of which has the name in argument.
   */

	dom *at(const char *);

 /**
    Selects an element according to these arguments:
    - the attribute,
    - the value of the attribute
    Returns the instance.
    Sets up the index to the right node, and error to false;
    if not found error is true;
   */

	dom *at(const char *, const char *);
	inline dom *at(text a, const char *b) { return at(a.data, b); }
	inline dom *at(text a, text b) { return at(a.data, b.data); }
	inline dom *at(const char *a, text b) { return at(a, b.data); }

 /**
    Compare the attribute and the value in parameters
    for the currently indexed node.
    Return true or false.
   */

	bool compare(const char *, const char *);
	inline bool hasValue(const char *a, const char *b) { return compare(a,b); }

 /**
   Get the value of an attribute.
   The attribute 'element' must point out the element of the node to read
   Returns:
   - the content if the parameter is "data" or
   - the value of the attribute that has the name in parameter or
   - an empty text if name not found.
   */

	const char *getValue(const char *);

 /**
    Get the data
    Input: index points out the node for the element.
   */

	const char *getData();


 /**
    Set a value
    Input: name of the attribute, new value
   */

	void setValue(const char *, const char *);

 /**
    Set the data
    Input: the text to assign as content of an element
   */

	void setData(const char *);

//========================= READING

/**
  Get the child of current node
  */

	int getChild();

/**
  Get the next node
  */

	int getNext();

/**
  Returns true if the current node is an xelement
  */

	bool isElement();
	bool isScript();
	bool isComment();


//========================= ITERATOR

 /**
   Going to the first element in the tree
   */

	dom *reset();

 /**
   Going to the first �l�ment at level
   */

	dom *begin();

 /**
    Points out the last node at the level.
    Set the error flag if no element preceeds at the level.
   */

	dom *end();

 /**
    Skip current node, points out the next one at the level.
    Set the error flag if no element follow at the level.
   */

	inline dom *next() { return inc(); }
	dom *inc();

 /**
    Points out the previous node at the level.
    Set the error flag if no element preceeds at the level.
   */

	dom *dec();

 /**
    Points out the first child of the current element
   */

	dom *down();

 /**
   Set the iterator index to a node
   */

	dom *setIndex(int);

 /**
   Get the current position in the iterator
   */

	inline int getIndex() { return index; }


 /**
    SetParent
    arguments: none
    When a tag is closed or at user call,
    gets the successor of the parent
   */
	dom *up();


//========================= STRUCTURE CHANGE


 /**
   Get Node
   argument: none
   Returns a pointer on the node pointed out inside a tree.
   */

	xnode *getNode();

 /**
   Get Element
   argument: none
   returns a pointer on the element pointed out inside the tree.
   If the node is not an element, return an empty element.
   */

	xelement *getElement();
	xscript  *getScript();
	xcomment *getComment();

 /**
   SetChild
   argument: a node or element (inside a xml document)
   Sets the node the child of the pointed out node, and
   links the current child as next node to the new node.
   */

	void setChild(xnode *);
	void setChild(xelement *);

 /**
   SetNext
   argument: a node or element (inside a xml document)
   Sets the node the successor of the pointed out node, and
   links the current successor as next node to the new node.
   */

	void setNext(xnode *);
	void setNext(xelement *);

 /**
   Remove
   Suppress the pointed out node.
   Childs are removed also.
   If it has a successor, it is linked to the previous node,
   but it there was no previous node, it becomes child to the parent.
   */

	void remove();

/**
  Push
  Add a node as last node inside the top node.
  The node and all sub-nodes are copied from a source
  Argument: source node (can have sub-nodes).
  */

	void push(xnode *);

/**
  Push Next
  Add a node as successor of currently pointed out node (by index).
  The node and sub-nodes from the source are copied into the tree.
  Argument: source node (can have sub-nodes).
  */

	void pushNext(xnode *);

 /**
  Push Child
  Add a node as child of currently pointed out node (by index).
  The node and sub-nodes from the source are copied into the tree.
  Argument: source node (can have sub-nodes).
  */

	void pushChild(xnode *);


//========================= XML FILE

 /**
    Save the xml source into a real xml file
   */

	void save(const char *, int = XML);

 /**
    Append the xml document from tree to an existing xml file
   */

	void append(const char *, int = XML);

 /**
   Load a xml document into the tree. Append.
   0 :  libxml2
   1 :  expat
   */

 //bool load(const char *, int = 0);


 /**
   Load a xml document into the tree.
   The first argument is a function (two ones are provided by extensions).
   SaxImport (default, for Sax)
   LXImport  (for libxml2)
   The second argument is the name of the xml file.
   */

	bool load(bool (*) (const char *), const char *);
	bool load(const char *);     // use default SaxImport


// INTERNALS

	void copyNext(xnode *);
	void copyChild(xnode *);
	void addTree(xnode *);
	void unknown(const char *);
	void noElement(const char *);

	inline void operator=(const char *b) { setData(b); }

};




extern dom *lxdom;
