/*
The Scriptol library
Home page of the compiler: https://www.scriptol.com

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/



#if !defined(_DYN_HEADER)
#define _DYN_HEADER


class item;
class dyn;
class sol_pair;
class solarray;
class dict;
class text;

class object
{
  public:
  object() { }
};


std::ostream& operator<< (std::ostream& out, dyn);
std::ostream& operator<< (std::ostream& out, dyn *);

//extern dyn Nil;

class nil {
  public:
  int flag;

  union {
   bool b;        // bool for error
   double n;
   text *t;
   solarray *a;
   dict *d;
   sol_pair *p;
   FILE *f;
   void *o;
   natural l;
  };

  nil();
  ~nil() {}

  bool is();
  static bool is(bool);
  static bool is(int);
  static bool is(natural);
  static bool is(double);
  static bool is(text &);
  static bool is(solarray &);
  static bool is(dict &);
  static bool is(dyn &);
  static bool is(FILE *);
  static bool is(object *);


};


/**
  C++ implementation of dynamic variables, with Scriptol's methods.
  */

class dyn : public nil  {
  public:
	  
  static dyn error;

  void set(unsigned char a);
  void set(char);
  void set(bool);
  void set(int);
  void set(unsigned int);
  void set(natural);
  void set(float);
  void set(double);
  void set(const char *);
  void set(const text &);
  void set(text *);
  void set(const dyn &);
  void set(dyn *);
  void set(sol_pair *);
  void set(solarray *);
  void set(const solarray &);
  void set(dict *);
  void set(const dict &);
  
  void set(FILE *);
  void set(void *);
  void set(const _arr32 &);
  void set(const _array64 &);
  void set(const _intArr &);
  void set(std::vector<dyn>);

  dyn()  { flag = 0; b = false; }

  /**
    Constructor: a dyn mays be created from a scalar, another
    dyn, an object or an solarray.
    */

  dyn(const dyn &a);

  inline dyn(unsigned char a) { set(a); }
  inline dyn(char a) { set(a); }
  inline dyn(bool a) { set(a); }
  inline dyn(int a) { set(a); }
  inline dyn(unsigned int a) { set(a); }
  inline dyn(natural a) { set(a); }
  inline dyn(float a) { set(a); }
  inline dyn(double a) { set(a); }
  inline dyn(const char *a) { set(a); }
  inline dyn(const text &a) { set(a); }
  inline dyn(text *a) { set(a); }
  inline dyn(dyn *a) { set(a); }
  inline dyn(sol_pair *a) { set(a); }
  inline dyn(solarray *a) { set(a); }
  inline dyn(const solarray &a) { set(a); }
  inline dyn(dict *a) { set(a); }
  inline dyn(const dict &a) { set(a); }
  inline dyn(FILE *a) { set(a); }
  inline dyn(void *a) { set(a); }
  inline dyn(_intArr *a) { set(a); }
  inline dyn(_arr32 *a) { set(a); }
  inline dyn(_array64 *a) { set(a); }
  
  inline dyn(std::vector<dyn> a) { set(a); }


  inline bool isPair()    { return flag == TYPE_PAIR; }
  inline bool isInt()     { return flag == TYPE_INT; }

  /**
    Returns true if the dyn is an integer.
    */
  inline bool isInteger() { return flag == TYPE_INT;  }

  /**
    Returns true if the dyn is a natural.
    */
  inline bool isNatural() { return isNumber(); }
  inline bool isReal() { return isNumber(); }  

  /**
    Returns true if the dyn is a number.
    */
  bool isNumber();

  /**
    Returns true if the dyn is a text.
    */
  inline bool isText()    { return flag == TYPE_TEXT; }

  /**
    Returns true if the dyn is an mixed solarray.
    */
  inline bool isArray()   { return flag == TYPE_ARRAY; }

  /**
    Returns true if the dyn is an mixed solarray.
    */
  inline bool isDict()   { return flag == TYPE_DICT; }


  /**
    Returns true if the dyn is a boolean.
    */
  inline bool isBoolean() { return flag == TYPE_BOOL; }

  /**
    Returns true if the dyn is a file.
    */
  inline bool isFile()    { return flag == TYPE_FILE; }

  /**
    Returns true if the dyn is a typed solarray.
    */
  inline bool isList()    { return flag == TYPE_TARRAY; }
  
  /**
    Returns true if the dyn is an object
    */
  inline bool isObject()    { return flag == TYPE_OBJECT; }
  

  /**
    Static.
    Returns true if the parameter is a string of digits.
    */
  static bool isNumber(const char *);
  static bool isNumber(const text &t);
  static bool isNumber(text *t);

  /**
    Returns the type of the solarray if the dyn is a typed solarray.
    */
  int arrayType();


  /**
    Returns the len if the dyn is a text.
    If it is a number, returns the number of bytes in memory.
    */
  int len();

  /**
    Return the size if the dyn is an solarray.
    */
  int size();

  bool     toBoolean();

  /**
    Converts the dyn to an integer.
    Return 0 if the dyn was not a number nor a text.
    */
  int      toInt();
  inline int toInteger() { return toInt(); }

  /**
    Converts the dyn to a natural.
    Return 0 if the dyn was not a number nor a text.
    */
  natural  toNatural();

  /**
    Converts the dyn to a real.
    Return 0 if the dyn was not a number nor a text.
    */
  double   toReal();

  /**
    Converts the dyn to a number.
    Return 0 if the dyn was not a number nor a text.
    */
  double   toNumber();
  sol_pair *toPair();

  /**
    Converts the dyn to a text.
    Return nil if the dyn can't be converted.
    Numbers, strings, solarray may be.
    */
  text     toText();      // any data

  /**
    Converts the dyn to C string of chars.
    Return nil if the dyn can't be converted.
    */
  const char *toStr();      // any data

  /**
    Converts the dyn to an solarray.
    Return an empty solarray if the dyn was not an solarray.
    */
  solarray    *toArray();

  /**
    Converts the dyn to a dict.
    Return an empty dict if the dyn was not a dict.
    */
  dict    *toDict();

  /**
    Converts the dyn to a file.
    Return null if the dyn was not a file.
    */
  FILE     *toFile();
    text     toFunction();

  /**
    Converts the dyn to an object.
    Return an empty object if the dyn was not an object.
    */
  void     *toObject();

  /**
    Returns an  solarray assigned to the dyn (default no parameter),
    but it mays also convert a scalar to an solarray solarray having its type.
    Ex: x.toList($(TYPE_TEXT)) builds an solarray with one element.
    */
  solarray *toList(int type = TYPE_ARRAY);

  /**
    Static.
    Converts the parameter to a boolean.
    */
  static bool toBoolean(dyn);

  /**
    Static.
    Converts the parameter to an integer.
    */
  static int      toInt(dyn);
  static inline int  toInteger(dyn d) { return toInt(d); }

  /**
    Static.
    Converts the parameter to a natural.
    */
  static natural  toNatural(dyn);

  /**
    Static.
    Converts the parameter to a real.
    */
  static double toReal(dyn);
  static double toReal(const char *);

  /**
    Static.
    Converts the parameter to a number.
    */
  static double   toNumber(dyn);

  /**
    Static.
    Converts the parameter to a string of chars.
    */
  static const char  *toStr(dyn);

  /**
    Static.
    Converts the parameter to an solarray.
    */
  static solarray  *toArray(dyn);

  /**
    Static.
    Converts the parameter to a dict.
    */
  static dict    *toDict(dyn);

  static sol_pair *toPair(dyn);

  /**
    Static.
    Converts the parameter to a text.
    */
  static text     toText(dyn);

  /**
    Static.
    Converts the parameter to a file.
    */
  static FILE     *toFile(dyn);

  /**
    Static.
    Converts the parameter to a typed solarray.
    */
  static void     *toList(dyn);

  bool equal(double, const char *);
  bool equal(double, text *);

  void clone(dyn);

  void operator=(dyn);
  template<class ANY> void operator=(ANY a) { set(a); }

  bool operator==(const dyn &);
  bool operator==(dyn *);
  bool operator==(int);
  bool operator==(natural);
  bool operator==(double);
  bool operator==(const char *);
  bool operator==(const text &);
  bool operator==(bool);
  bool operator==(const solarray &);
  bool operator==(const dict &);

  bool operator!=(int);
  bool operator!=(natural);
  bool operator!=(double);
  bool operator!=(const dyn &);
  bool operator!=(dyn *);
  bool operator!=(const char *);
  bool operator!=(const text &);
  bool operator!=(bool);
  bool operator!=(const solarray &);
  bool operator!=(const dict &);

  bool operator<(int);
  bool operator<(natural);
  bool operator<(double);
  bool operator<(const dyn &d);
  bool operator<(dyn *);
  bool operator<(const char *);
  bool operator<(const text &);
  bool operator<(bool);
  bool operator<(const solarray &);
  bool operator<(const dict &);

  bool operator>(int);
  bool operator>(natural);
  bool operator>(double);
  bool operator>(const dyn &d);
  bool operator>(dyn *);
  bool operator>(const char *);
  bool operator>(const text &);
  bool operator>(bool);
  bool operator>(const solarray &);
  bool operator>(const dict &);

  inline bool operator>=(int b) { return !operator<(b); }
  inline bool operator>=(double b) { return !operator<(b); }
  inline bool operator>=(bool b) { return !operator<(b); }
  inline bool operator>=(const text &b) { return !operator<(b); }
  inline bool operator>=(const char *b) { return !operator<(b); }
  inline bool operator>=(const solarray &b) { return !operator<(b); }
  inline bool operator>=(const dict &b) { return !operator<(b); }
  inline bool operator>=(const dyn &b) { return !operator<(b); }
  inline bool operator>=(dyn *b) { return !operator<(b); }

  inline bool operator<=(int b) { return !operator>(b); }
  inline bool operator<=(double b) { return !operator>(b); }
  inline bool operator<=(bool b) { return !operator>(b); }
  inline bool operator<=(const text &b) { return !operator>(b); }
  inline bool operator<=(const char *b) { return !operator>(b); }
  inline bool operator<=(const solarray &b) { return !operator>(b); }
  inline bool operator<=(const dict &b) { return !operator>(b); }
  inline bool operator<=(const dyn &b) { return !operator>(b); }
  inline bool operator<=(dyn *b) { return !operator>(b); }


  double operator+(int);
  double operator+(natural);
  double operator+(double);
  const char *operator+(const char *);
  text operator+(const text &);
  dyn operator+(const dyn &);
  solarray operator+(const solarray &);
  solarray operator+(solarray *);
  dict operator+(const dict &);
  dict operator+(dict *);


  double operator-(int);
  double operator-(natural);
  double operator-(double);
  dyn operator-(const dyn &);
  solarray operator-(const solarray &);
  dict operator-(const dict &);

  double operator*(int);
  double operator*(natural);
  double operator*(double);
  double operator*(dyn);

  double operator/(int);
  double operator/(natural);
  double operator/(double);
  double operator/(dyn);

  double operator%(int);
  double operator%(natural);
  double operator%(double);
  double operator%(dyn);

  int  operator&(dyn);
  int  operator&(int);
  int  operator&(double);

  int  operator|(dyn);
  int  operator|(int);
  int  operator|(double);

  bool operator&&(dyn);
  bool operator&&(int);
  bool operator&&(double);

  bool operator||(dyn);
  bool operator||(int);
  bool operator||(double);


  void operator+=(dyn);
  void operator+=(int);
  void operator+=(natural);
  void operator+=(double);
  void operator+=(const char *);
  void operator+=(const text &);
  void operator+=(const solarray &);
  void operator+=(const dict &);

  void operator-=(dyn);
  void operator-=(int);
  void operator-=(natural);
  void operator-=(double);
  void operator-=(const solarray &);
  void operator-=(const dict &);

  void operator*=(dyn);
  void operator*=(int);
  void operator*=(natural);
  void operator*=(double);

  void operator/=(dyn);
  void operator/=(int);
  void operator/=(natural);
  void operator/=(double);

  void operator<<=(int);
  void operator>>=(int);

  bool operator!();
  solarray &operator,(dyn);

  dyn at(int);
  inline dyn at(natural a) { return at( (int) a); }
  dyn at(const text &);
  dyn at(const char *);
  
  
  dyn  operator[](int);
  inline dyn operator[](natural a) { return operator[]((int) a); }
  dyn  operator[](const text &);
  dyn  operator[](const char *);
  dyn  operator[](dyn);
  
  
  void setAt(int, dyn);
  inline void setAt(natural a, dyn b) { setAt((int) a, b); }
  void setAt(int, const text &);
  inline void setAt(natural a, const text &b) { setAt((int) a, b); }
  void setAt(int, const char *);
  inline void setAt(natural a, const char *b) { setAt((int) a, b); }

  void setNil() { setFalse(); }
  static bool isNil(int a) { return a == 0; }
  static bool isNil(natural a) { return a == 0; }
  static bool isNil(dyn);
  bool isNil();

  static dyn dummyNil;
  void setFalse() { flag = TYPE_BOOL; b = 0; }


  //===== text

  text trim();
  text lTrim();
  text rTrim();
  text lower();
  text upper();

  void cat(char *);
  void cat(dyn);
  void cat(const text &);

  //===== solarray: iterators
  void reset();
  dyn  inc();
  dyn  dec();
  dyn  value();
  text key();
  dyn  begin();
  dyn  end();
  int index();

  //==== stack
  void push(dyn);
  dyn pop();
  dyn shift();
  void unshift(dyn);

  solarray slice(int, int, int);
  void splice(int, int, solarray, int);
  solarray slice(natural, natural, natural);
  void splice(natural, natural, solarray, natural);
  double sum();

  //=====
  void display();
  static text str(dyn);

  int nomatch(char);
  void boolToNum();
};


text operator+(text, dyn);
text operator+(const char *, dyn);
dyn operator+(int , dyn);
dyn operator+(double , dyn);
dyn operator+(solarray *, dyn);
dyn operator+(dict *, dyn);

dyn operator-(int , dyn);
dyn operator-(double , dyn);
dyn operator-(solarray *, dyn);
dyn operator-(dict *, dyn);

dyn operator*(int , dyn);
dyn operator*(double , dyn);

dyn operator/(int , dyn);
dyn operator/(double , dyn);

int operator&(int , dyn);
int operator&(double , dyn);

int operator|(int , dyn);
int operator|(double , dyn);


inline bool operator==(int a, dyn b)         { return b == a; }
inline bool operator==(double a, dyn b)      { return b == a; }
inline bool operator==(const text &t, dyn d) { return d == t; }
inline bool operator==(bool b, dyn d)        { return d == b; }
inline bool operator==(const char *t, dyn d) { return d == t; }

inline bool operator!=(int a, dyn b)         { return b != a; }
inline bool operator!=(double a, dyn b)      { return b != a; }
inline bool operator!=(const text &t, dyn d) { return d != t; }
inline bool operator!=(bool b, dyn d)        { return d != b; }
inline bool operator!=(const char *t, dyn d) { return d != t; }

inline bool operator<=(int a, dyn b)         { return b > a; }
inline bool operator<=(double a, dyn b)      { return b > a; }
inline bool operator<=(const text &t, dyn d) { return d >  t; }
inline bool operator<=(const char *t, dyn d) { return d >  t; }

inline bool operator>=(int a, dyn b)         { return b < a; }
inline bool operator>=(double a, dyn b)      { return b < a; }
inline bool operator>=(const text &t, dyn d) { return d >  t; }
inline bool operator>=(const char *t, dyn d) { return d <  t; }

inline bool operator>(int i, dyn b)          { return b.operator<=(i); }
inline bool operator>(double d, dyn b)       { return b.operator<=(d); }
inline bool operator>(const text &t, dyn d)  { return d.operator<=(t); }
inline bool operator>(const char *t, dyn d)  { return d.operator>=(t); }

inline bool operator<(int i, dyn b)          { return b.operator>=(i); }
inline bool operator<(double d, dyn b)       { return b.operator>=(d); }
inline bool operator<(const text &t, dyn d)  { return d.operator>=(t); }
inline bool operator<(const char *t, dyn d)  { return d.operator<=(t); }


void operator+=(int, dyn);
void operator+=(double, dyn);
void operator+=(text, dyn);
void operator+=(char *, dyn);

void operator-=(int, dyn);
void operator-=(double, dyn);

void operator*=(int, dyn);
void operator*=(double, dyn);

void operator/=(int, dyn);
void operator/=(double, dyn);

void operator&=(int, dyn);
void operator&=(double, dyn);

void operator|=(int, dyn);
void operator|=(double, dyn);

bool operator&&(int, dyn);
bool operator&&(double, dyn);

bool operator||(int, dyn);
bool operator||(double, dyn);


#endif