/*
The Scriptol library
Home page of the compiler: https://www.scriptol.com

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/



#if !defined(_SOL_FUN)
#define _SOL_FUN

#define STR_PAD_LEFT 0
#define STR_PAD_RIGHT 1
#define STR_PAD_BOTH 2


//====== internal, replace dirent.h for all including visual c

typedef int DIR;

struct dirent {
    int d_ino;
    char *d_name;
};

DIR*            opendir (const char*);
struct dirent*  readdir (DIR*);
int             closedir (DIR*);
void            rewinddir (DIR*);
long            telldir (DIR*);
void            seekdir (DIR*, long);

//===== general

/**
  Displaying a solarray
  */

#define print_r(a) a.display()

const char *plural(int);

void runtimeError(const char *);

/**
  Displaying a message before to exit
  */
void die(text);
void die(const char *);

/**
  Stop the program, return to operating system
  */
void exit();

/**
  Swap the content of two variables.
  */
#define swap(a,b) { dyn temp = a; a = b; b = temp; }

/**
  Get the minimum of two values
  */
int min(int, int);
natural min(natural, natural);
double min(double, double);
dyn min(dyn, dyn);
text min(text, text);

/*
  Get the sign of a value : -1, 0, 1
  */
int _sol_sgn(double);
int _sol_sgn(int);


/**
  Get the maximum of two values
  */
int max(int, int);
natural max(natural, natural);
double max(double, double);
dyn max(dyn, dyn);
text max(text, text);

//===== types

/**
  Return true if the argument is a dict
  */
bool is_associative(dyn a);

/**
  Return true if the argument is a typed array.
  */
bool is_list(dyn a);

//===== numbers

/**
  Convert the argument to natural and return it.
  The argument must be a scalar.
  */
natural naturalval(dyn);
natural naturalval(unsigned char);
natural naturalval(int);
natural naturalval(natural);
natural naturalval(float);
natural naturalval(double);
natural naturalval(text);
natural naturalval(const char *);
natural naturalval(void *);

/**
  Convert the argument to number and return it.
  The argument must be a scalar.
  */
double doubleval(dyn);
double doubleval(text);
double doubleval(const char *);
double doubleval(int);
double doubleval(natural);
double doubleval(double);
double doubleval(void *);

/**
  Convert the argument to integer and return it.
  The argument must be a scalar.
  */
int intval(dyn);
int intval(text);
int intval(const char *);
int intval(double);
int intval(int);
int intval(natural);
int intval(void *);

/**
  Convert the argument to a C string of char and return it.
  The argument must be a scalar.
  */
char *str(bool);  
char *str(int);
char *str(natural);
char *str(double);
const char *str(dyn);
char *str(void *);
char *str(solarray);
char *str(text);
char *str(char *);

/**
  Convert the argument to a text and return it.
  The argument must be a scalar.
  */
text strval(int);
text strval(natural);
text strval(double);
text strval(dyn);
text strval(void *);
text strval(bool);
text strval(text);
text strval(char *);

solarray arrayval(_arr32);
solarray arrayval(_intArr);
solarray arrayval(_array64);

/**
  Convert the int to bool
  */

bool boolval(int);

/**
  Return the text holding an integer as a pure integer.
  The argument must be a text.
  */
int ord(const char *);
int ord(text);
int ord(char);

/**
  Return the integer argument as a text.
  */
text chr(int);

double _sol_round(double);

//===== solarray

#define array_values(a) a.values()
#define array_keys(a) a.keys()
#define array_flip(a) a.flip()

int   count(solarray);

solarray _sol_range(int, int);

solarray array_merge(solarray, solarray);
solarray array_diff(solarray , solarray);
solarray array_intersect(solarray, solarray);
solarray array_unique(const solarray &);
int array_search(solarray, ...);
solarray array_reverse(solarray, bool);
char * _sol_dumpArray(solarray);

text implode(solarray *, const char *);

void array_map(void (*)(dyn &), solarray &);
void array_map(void (*)(dyn *), solarray &);
void array_map(void (*fun)(dyn &, dyn &), solarray &, solarray &);
void array_map(void (*fun)(dyn &, dyn &, dyn &), solarray &, solarray &, solarray &);


bool in_array(int, solarray);
bool in_array(text, solarray);
bool in_array(const char *, solarray);
bool in_array(double, solarray);
bool in_array(dyn, solarray);

bool in_sequence(const char *, text &);
bool in_sequence(const char *, const char *);


//===== iterators

dyn reset(solarray);
dyn next(solarray);
dyn prev(solarray);
dyn end(solarray);
text key(solarray &);
dyn current(solarray &);

//===== stack

#define array_unshift(a, d) a.unshift(d)
#define array_shift(a) a.shift()
#define array_pop(a) a.pop()
#define array_push(a, d)  a.push(d)

//===== functions

text implode(solarray, const char *);

//===== file

solarray file(text);
solarray file(const char *);
void store(solarray, const char *);
void store(solarray, text);
void require(const char *);
void require_once(const char *);

//===== text

char *builtin_hex(int);

char *strtoupper(const char *);
char *strtolower(const char *);
char *ucfirst(const char *);
char *ucwords(const char *);
char *wordwrap(const char *, int, const char *);

int strlen(text);
int strlen(dyn);
int strcasecmp(text , text );
#ifdef _UNIX
int strcasecmp(const char *, const char *);
#endif

void str_replace(text , text );

char *substr(const char *, int, int);
char *substr(const char *, int);
int substr_count(const char *, const char *);
char *substr_replace(const char *, const char *, int, int);
char *stristr(const char *, const char *);
char *str_replace(const char *, const char *, const char *);
char *strtr(const char *, const char *, const char *);
char *str_repeat(const char *, int);
int strpos(const char *, const char *, int);

solarray explode(const char *, const char *);

char *chop(const char *);
char *rtrim(const char *);
char *ltrim(const char *);
char *trim(const char *);

text str_pad(text , int, text = " ", int = STR_PAD_RIGHT);
text str_pad(text , int, const char * =  " ", int = STR_PAD_RIGHT);
text str_pad(const char *, int, const char * = " ", int = STR_PAD_RIGHT);

//===== file

FILE *fopen(text, text);
FILE *fopen(text, const char *);
FILE *_sol_open(const char *, const char *);
void fseek(FILE *, int);
inline void fputs(const text &t) { fputs(t.data); }
text fgets(FILE *);
char *fread(FILE *, int size);
int  fwrite(FILE *, const text &);
int  fwrite(FILE *, const char *);
char *_sol_read(FILE *, int);
int _sol_write(FILE *, char *, int);
solarray _sol_ftoarr(char *);
text _sol_ftostr(char *);

int binOpen(const char *name, const char *mode);
char *binRead(int handle, int size);
int binWrite(int handle, char *data, int size);
int binClose(int handle);

//===== dir

DIR * _ScriptolDir(const char *);

bool file_exists(const text &);
bool file_exists(const char *);
text filetype(const char *);
double filesize(const char *);
int  filemtime(const char *);
const char *date(const char *, int);
bool _sol_chdir(const char *);              // change directory for a drive/current
bool _sol_mkdir(const char *);              // create directory
bool _sol_rmdir(const char *);              // remove directory
bool _sol_ren(const char *, const char *);  // rename file

bool _sol_isdir(const char *);
bool _sol_isfile(const char *);
bool _sol_unlink(const char *);

DIR *open_dir(const char *);
text read_dir(DIR *);
void rewind_dir(DIR *);       // removed for now (not useful)
void close_dir(DIR *);
text getcwd();

solarray scandir(const char *);

//===== html

char *extractProperty(const char *, const char *);
dict get_meta_tags(const char *);

//===== regular not yet implemented

bool ereg(text t, text);
bool ereg(text t, const char *);
bool ereg(const char *, text);
bool ereg(const char *, const char *);


//===== Type in dyn

bool is_dyn_array(dyn);
bool is_dyn_list(dyn);
bool is_dyn_object(dyn);
bool is_dyn_dict(dyn);

/**
  Input: gets a string from the keyboard. Max length 255 chars
  Argument: optional, a text to display
  Returns:  the string entered
  */

char *input(const char * = "");


//===== Internals

char * _sol_inputText(const char * = "");
double _sol_inputNumber(const char * = "");
char *_sol_phpstring(const char *, ... );
const char *_sol_phpcvt(text *);
const char *_sol_phpcvt(dyn *);



#endif



