/*
The Scriptol library
Home page of the compiler: https://www.scriptol.com

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/


#if !defined(_SOL_TEXT)
#define _SOL_TEXT

/*
#ifdef UNIX
typedef unsigned long long natural;
#else
typedef unsigned __int64 natural;
#endif
*/
typedef uint64_t natural;

/**
  Basical builtin text object
  */


class text 
{
 public:

 /**
   Dynamic size of the string
   */

 int len;
 char *data;

 text *op;

 text();

 /**
   Constructor from a C string
   */

 text(const char *);

 /**
   Constructor for the = operator
   */
 text(const text &);

 text(text *);

 text(char);

 /**
   Constructor: convert an integer into a string of digits
   */
 text(int);
 text(unsigned int);
 text(natural);

 /**
   Constructor: convert a number into a string of digits
   */

 text(double);
 text(bool);
 text(void *);
 ~text();

 /**
   Converts the text into a buffer, with size in parameter
   */

 void reserve(int);

 void set(text *);
 void set(const text &);
 void set(const char *);
 void set(int);
 void set(unsigned int);
 void set(natural);
 void set(double);
 void set(bool);
 void set(char);


 /**
   Returns the current size of the string
   */
 inline int length() { return len; }


 /**
   Returns true if the text is a null string
   */
 inline int empty()  { return len == 0; }


 /**
   Converts to upper case
   */
 text upper();


 /**
   Converts to lower case
   */
 text lower();


 /**
   Converts the first letter to upper case (other ones unchanged)
   */
 text capitalize();


 /**
   Concatenates the string with the one in parameter.
   */
 void cat(const char *);

 bool in(const char *);
 bool in(const text &);


 /**
   Searches for the given string inside the text.
   An optional second parameter is the starting position in the text.
   Don't ignore case.
   Returns the position of the substring found inside the text.
   */
 int find(const char *, int = 0);

 /**
   Searches for the last occurence of given string inside the text.
   An optional second parameter is the starting position in the text.
   Don't ignore case.
   Returns the position of the substring found inside the text.
   */
 int findLast(const char *, int = 0);


 /**
   Searches for the given string inside the text.
   An optional second parameter is the starting position in the text.
   Ignores case.
   */
 int findLex(const char *, int = 0);


 /**
   Static method.
   Compares two strings. Optionally limits the length of the strings,
   according to the third argument.
   Ignores case.
   Returns:
   -1   if the first is lesser than the second one.
   0    if the strings are identical.
   1    if the first is greater than the second one.
   */
 static int compareLex(const char *, const char *, int);

 /**
   Compares the text with a given strings.
   Ignores case.
   Returns:
   -1   if the text is lesser than the string.
   0    if the strings are identical.
   1    if the text is greater than the given string.
   */
 int compare(const char *);

 /**
   Compares the text with the given strings.
   Don't ignores case.
   Returns:
   -1   if the text is lesser than the string.
   0    if the strings are identical.
   1    if the text is greater than the given string.
   */
 int identical(const char *);

 /**
   Replaces inside the text, the first string by the second one.
   Don't ignores case.
   */
 void replace(const char *, const char *);

 /**
   Inserts a string inside the text at the given location (first argument).
   */
 void insert(int, const char *);

 text slice(int, int, int);
 const char * substr(int, int);      // as slice, return char pointer
 const char * substr(int);

 void splice(int, int, const char*, int);
 void splice(int a, int b, const text &t, int i);

 /**
   Removes trailing blank spaces and control characters.
   */
 text rTrim();

 /**
   Removes heading blank spaces and control characters.
   */
 text lTrim();

 /**
   Removes both heading and trailing blank spaces and control characters.
   */
 text trim();

 /**
   Builds a string from n occurences (the second parameter)
   of the string given as first parameter.
   */
 void fill(const char *, int);

 /**
   Builds a text from n occurences (the parameter) of the text,
   and returns it.
   */
 text dup(int);


 char toChar();

 /**
   Converts the text, if it is a string of digits, to an integer,
   and returns it.
   Returns 0 if no a number.
   */
 int toInt();

 /**
   As toInt().
   */
 int toInteger();

 /**
   Static function: converts a string into a number,
   return a double or 0 if not a string of digits.
   */
 static double toNumber(const char *);

 /**
   Converts the text into a number,
   returns a number or 0 if the text is not a string of digits.
   */
 double toNumber();

 /**
   Converts the text, if it is a string of digits, to a natural,
   and returns it.
   Returns 0 if no a number.
   */
 unsigned long toNatural();

 /**
   Converts the text into a real number,
   returns a real or 0 if the text is not a string of digits.
   */
 inline double toReal()     { return toNumber(); }

 /**
   Save the text into a file.
   */
void toFile(const char *);
 
 /**
   Returns the C string that the object holds.
   */
 char *toStr(); 

 /**
   Returns the C string that the object holds.
   */
 inline char *c_str() { return (char *) data; }

 /**
   Returns true if the text is a string of digits.
   */
 bool isNumber();

 /**
   Static function.
   Returns true if the given string is a string of digits.
   */
 static bool isNumber(const char *);

 bool operator==(const text &);
 bool operator==(const char *);
 bool operator==(int);
 bool operator==(double);
 bool operator==(bool);

 bool operator!=(const text &);
 bool operator!=(const char *);
 bool operator!=(int);
 bool operator!=(double);
 bool operator!=(bool);

 bool operator<=(const text &);
 bool operator<=(const char *);

 bool operator>=(const text &);
 bool operator>=(const char *);

 bool operator<(const text &);
 bool operator<(const char *);

 bool operator>(const text &);
 bool operator>(const char *);

 bool operator!();
 
 void operator=(const text &);
 void operator=(const char *);

 text operator+(const text &);
 text operator+(const char *);

 void operator+=(const text &);
 void operator+=(const char *);

 text &operator[](int);

 void setAt(int, const text &);
 text operator*(int);

};

std::ostream & operator<< (std::ostream &, text);

inline bool operator==(const char *a, text b) { return b.operator==(a); }
inline bool operator==(bool a, text b) { return b.operator==(a); }
inline bool operator!=(const char *a, text b) { return b.operator!=(a); }
inline bool operator!=(bool a, text b) { return b.operator!=(a); }
inline bool operator<=(const char *a, text b) { return b.operator>(a); }
inline bool operator>=(const char *a, text b) { return b.operator<(a); }
inline bool operator<(const char *a, text b) { return b.operator>(a); }
inline bool operator>(const char *a, text b) { return b.operator<(a); }
inline text operator+(const char *a, text b) { return text(a) + b; }


#endif
